/* pcl725io.c - Device driver for the Advantech PCL-725 and ADClone ACL-725
		relay actuator and isolated D/I cards */

/* Written 1996 by Werner Almesberger, EPFL LRC */


#include <linux/config.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/string.h>
#include <linux/kernel.h> /* printk */
#include <linux/fs.h>
#include <linux/ioport.h>
#include <asm/io.h>

#include "pcl725.h"


#ifndef MODULE
#error "PCL-725/ACL-725 driver must be compiled as a module"
#endif


static unsigned short pcl725_map[PCL725_MAX_DEVS];


static int pcl725_ioctl(struct inode *inode,struct file *filp,unsigned int cmd,
  unsigned long arg)
{
	unsigned short base;

	base = pcl725_map[MINOR(inode->i_rdev)];
	switch (cmd) {
		case PCL725_INP_GET:
			return inb_p(base+1);
		case PCL725_REL_GET:
			return inb_p(base);
		case PCL725_REL_SET:
			outb_p(arg,base);
			return 0;
		case PCL725_REL_ON:
			outb_p(inb_p(base) | arg,base);
			return 0;
		case PCL725_REL_OFF:
			outb_p(inb_p(base) & ~arg,base);
			return 0;
		case PCL725_REL_TOGGLE:
			outb_p(inb_p(base) ^ arg,base);
			return 0;
		default:
			return -EINVAL;
	}
}


static int pcl725_open(struct inode *inode,struct file *filp)
{
	if (MINOR(inode->i_rdev) >= PCL725_MAX_DEVS) return -EINVAL;
	if (!pcl725_map[MINOR(inode->i_rdev)]) return -ENODEV;
	MOD_INC_USE_COUNT;
	return 0;
}


static void pcl725_release(struct inode *inode,struct file *filp)
{
	MOD_DEC_USE_COUNT;
}


static struct file_operations pcl725_fops = {
	NULL,		/* lseek */
	NULL,		/* read */
	NULL,		/* write */
	NULL,		/* readdir */
	NULL,		/* select */
	pcl725_ioctl,	/* ioctl */
	NULL,		/* mmap */
	pcl725_open,	/* open */
	pcl725_release,	/* release */
	NULL		/* fsync */
};


static int io[PCL725_ADDRS];


int init_module(void)
{
	int i,error,index;

	if (register_chrdev(PCL725_MAJOR,DEV_NAME,&pcl725_fops)) {
		printk("unable to get major %d for PCL-725/ACL-725 devs\n",
		    PCL725_MAJOR);
		return -EIO;
	}
	memset(pcl725_map,0,sizeof(pcl725_map));
	error = 0;
	for (i = 0; i < PCL725_ADDRS; i++)
		if (io[i]) break;
	if (i == PCL725_ADDRS) io[0] = PCL725_DFL_ADDR;
	index = 0;
	for (i = 0; i < PCL725_MAX_DEVS; i++) 
		if (io[i]) {
			if (io[i] < PCL725_BASE || io[i] >= PCL725_BASE+
			    (PCL725_ADDRS << 1) || (io[i] & 1)) {
				printk(DEV_NAME ": invalid address 0x%x\n",
				    io[i]);
				error = -EINVAL;
				break;
			}
			if (index == PCL725_MAX_DEVS) {
				printk(DEV_NAME ": too many devices. "
				    "Increase PCL725_MAX_DEVS and recompile."
				    "\n");
				break;
			}
			error = check_region(io[i],PCL725_PORTS);
			if (error) break;
			request_region(io[i],PCL725_PORTS,DEV_NAME);
			printk(DEV_NAME "%d: registered at address 0x%x\n",
			    index,io[i]);
			pcl725_map[index++] = io[i];
		}
	if (!error) return 0;
	while (--index >= 0) release_region(pcl725_map[index],PCL725_PORTS);
	return error;
}


void cleanup_module(void)
{
	int i;

	for (i = 0; i < PCL725_MAX_DEVS; i++)
		if (pcl725_map[i])
			release_region(pcl725_map[i],PCL725_PORTS);
	if (unregister_chrdev(PCL725_MAJOR,DEV_NAME))
		printk("unable to unregister major %d for PCL-725/ACL-725 "
		    "devs\n",PCL725_MAJOR);
}
